/*
 * gc global variable provides access to GUI Composer infrastructure components and project information.
 * For more information, please see the Working with Javascript guide in the online help.
 */
var gc = gc || {};
gc.services = gc.services || {};

/*
 *  Boilerplate code for encoding and decoding data transmitted to / received from TI devices.
 */

var targetAccessErrorHandler = function (error) {
  if (error && error.message) {
    throw error.message;
  } else {
    throw error;
  }
};

var count = 0;

/**
 * This code implements the IPacketCodec interface.
 *
 * @constructor
 * @implements gc.databind.IPacketCodec
 */
CustomCodec = function () {};

// use default implementations of connect,disconnect,encode and decode functions:
CustomCodec.prototype = new gc.databind.IPollingPacketCodec();

CustomCodec.prototype.writeCmdTimer = null;
CustomCodec.prototype.maxBitWidth = 16;
CustomCodec.prototype.readCmdSymbols = null;
CustomCodec.prototype.writeCmdSymbols = null;
CustomCodec.prototype.pendingWriteCmds = null;
CustomCodec.prototype.refreshListener = null;
CustomCodec.prototype.preRefreshListener = null;
CustomCodec.prototype.skipReadsUntilNextReadCmdFlagSet = false;
CustomCodec.prototype.preRefreshTimeInMs = 30;
CustomCodec.prototype.regBlocks = null;

var cmdInProgress = false;
var isNVMInProgress = false;
var commQueue = [];
var memoryInfo;
var intGlobalVariableDataAddress = {};
var registerProcessErrorCount = 0;

var processRegisterProcessError = function (thisCmd) {
  registerProcessErrorCount++;
  if (registerProcessErrorCount === 5) {
    templateObj.$.ti_connectivity_issue_dialog.open();
    return;
  }
  setTimeout(() => {
    commQueue.push(thisCmd);
  }, 1000);
}

var getArrayGlobalVariable = function (variableName) {
  return gc.target.access.readValue(variableName);
}

var getIntGlobalVariable = function (variableName) {
  return gc.target.access.readValue(variableName);
}

var setArrayGlobalVariable = function (variableName, value) {
  return gc.target.access.writeValue(variableName, value);
}

var setIntGlobalVariable = function (variableName, value) {
  if (intGlobalVariableDataAddress[variableName]) {
    return gc.target.access.writeMemory(intGlobalVariableDataAddress[variableName], undefined, [value])
  }

  return gc.target.access.ready.then(dsLite => {
    return dsLite.expressions.evaluate(`&${variableName}`)
  }).then(data => {
    intGlobalVariableDataAddress[variableName] = +data.value;
    return gc.target.access.writeMemory(intGlobalVariableDataAddress[variableName], undefined, [value])
  });
}

setInterval(function () {
  if (cmdInProgress || isNVMInProgress || commQueue.length == 0) {
    return;
  }

  cmdInProgress = true;
  var thisCmd = commQueue.shift();
  if (thisCmd.processType === 'R') {
    getIntGlobalVariable('MEMADDR_PROG').then((value) => {
      if (value === 2) {
        return value;
      } else {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error(`Unexpected Error: The resolved value is ${value}`);
        });
      }
    }).then(value => {
      return setArrayGlobalVariable('DATAWR_ARRAY', thisCmd.packet);
    }).then(value => {
      return setIntGlobalVariable('MEMADDR_PROG', 7);
    }).then(value => {
      return getIntGlobalVariable('MEMADDR_PROG');
    }).then(value => {
      if (value === 10) {
        return value;
      } else if (value === 0) {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error('Error in Sent Packet: Rewrite Packet');
        })
      } else {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error(`Unexpected Error: The resolved value is ${value}`);
        });
      }
    }).then(value => {
      return getArrayGlobalVariable('DATARD_ARRAY');
    }).then(arrayValue => {
      let result = deserializeCommandPacket(thisCmd, arrayValue, true);
      if (result != -1) {
        thisCmd.promise.resolve(result);
        return setIntGlobalVariable('MEMADDR_PROG', 2);
      } else {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error('Error in Received Packet: Rewrite Packet');
        });
      }
    }).then(() => {
      cmdInProgress = false;
      console.log(`${thisCmd.symbol} is resolved for read request`);
    }).fail(function (err) {
      thisCmd.promise.reject(err);
      console.log(`${thisCmd.symbol} is rejected for read request - ${err instanceof Error ? err : err instanceof Object ? JSON.stringify(err) : err}`);
      cmdInProgress = false;
    });
  } else if (thisCmd.processType === 'W') {
    getIntGlobalVariable('MEMADDR_PROG').then((value) => {
      if (value === 2) {
        return value;
      } else {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error(`Unexpected Error: The resolved value is ${value}`);
        });
      }
    }).then(value => {
      return setArrayGlobalVariable('DATAWR_ARRAY', thisCmd.packet);
    }).then(value => {
      thisCmd.promise.resolve();
      return setIntGlobalVariable('MEMADDR_PROG', 7);
    }).then(value => {
      return getIntGlobalVariable('MEMADDR_PROG');
    }).then((value) => {
      if (value == 2) {
        console.log(`${thisCmd.symbol} is resolved for write request`);
        cmdInProgress = false;
      } else if (value === 0) {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error('Error in Sent Packet: Rewrite Packet');
        })
      } else {
        return setIntGlobalVariable('MEMADDR_PROG', 2).then(() => {
          processRegisterProcessError(thisCmd);
          throw new Error(`Unexpected Error: The resolved value is ${value}`);
        });
      }
    }).fail(function (err) {
      thisCmd.promise.reject(err);
      console.log(`${thisCmd.symbol} is rejected for write request: ${err instanceof Error ? err : err instanceof Object ? JSON.stringify(err) : err}`);
      cmdInProgress = false;
    });
  }
}, 10);

function updateProcessQueue(packet, processType, symbol, size) {
  var deferred = Q.defer();
  commQueue.push({
    processType: processType,
    promise: deferred,
    packet: packet,
    symbol: symbol,
    size: size
  });
  return deferred.promise;
}

function valueToBytesArray(value, bit_size, endian) {
  var buf = [];
  if (bit_size <= 8) {
    buf.push(value);
  } else {
    var bs = (((bit_size - 1) >> 3) << 3);
    if (endian == 'big') {
      for (var b = bs; b >= 0; b -= 8) {
        buf.push(Number(BigInt(value) >> BigInt(b)) & 0xff);
      }
    } else { // (endian == 'little')
      for (var b = 0; b <= bs; b += 8) {
        buf.push(Number(BigInt(value) >> BigInt(b)) & 0xff);
      }
    }
  }
  return buf;
}

function bytesArrToValue(arr, byteSize, endian) {
  if (byteSize == 64) {
    let arrayCollection = [];
    arrayCollection.push(arr);
    arrayCollection.push(arr.splice(4, 8));

    if (endian == 'little') {
      arrayCollection[0] = arrayCollection[0].reverse();
      arrayCollection[1] = arrayCollection[1].reverse();
    }
    return arrayCollection.map((arr) => {
      return '0x' + arr.map((data) => {
        return ("0" + (Number(data).toString(16))).slice(-2)
      }).join('')
    });
  } else {
    if (endian == 'little') {
      arr = arr.reverse();
    }
    return '0x' + arr.map((data) => {
      return ("0" + (Number(data).toString(16))).slice(-2)
    }).join('');
  }
}

CustomCodec.prototype.writeValue = function (symbol, value) {
  let packet = this.createCommandPacket(symbol, false, true, value);
  return updateProcessQueue(packet, 'W', symbol.name, symbol.size);
};

CustomCodec.prototype.readValue = function (symbol) {
  let packet = this.createCommandPacket(symbol, true, true);
  return updateProcessQueue(packet, 'R', symbol.name, symbol.size);

};

CustomCodec.prototype.onPreRefresh = function (oldValue, newValue, progress) {
  if (this.pendingWriteCmds && this.pendingWriteCmds.length > 0) {
    gc.console.log("CustomCodec", "skipping setting readCmd flag because of pending write(s)")
  } else {
    gc.console.debug("CustomCodec", "preRefresh: set ReadCmd flag");
  }
};

CustomCodec.prototype.onRefresh = function (oldValue, newValue, progress) {
  gc.console.log("CustomCodec", "refresh");
};

CustomCodec.prototype.createCommandPacket = function (regInfo, isRead, isCrcEnabled, writeValue) {
  var addr = regInfo.addr;
  var slaveAddr = 192;
  var memoryAddr = (addr & 0x000000FF).toString(2).padStart(12, "0");
  var memorySection = parseInt(memoryInfo[(addr & 0xFFFF0000).toString(16).padStart(8, "0").slice(0, 4)].memorySection).toString(2).padStart(4, "0");
  var memoryPage = parseInt(memoryInfo[(addr & 0xFFFF0000).toString(16).padStart(8, "0").slice(0, 4)].memoryPageObj[(addr & 0x0000F000).toString(16).slice(0, 1)].memoryPage).toString(2).padStart(4, "0");
  var remainingBits = ((isRead) ? "1" : "0") + ((isCrcEnabled) ? "1" : "0") + ((regInfo.size == 16) ? "00" : ((regInfo.size == 32) ? "01" : "10"));
  var commandPacket = [slaveAddr];
  commandPacket = commandPacket.concat(valueToBytesArray(parseInt(remainingBits + memorySection + memoryPage + memoryAddr, 2), 24, 'big'));

  if (isRead) {
    return commandPacket;
  } else {
    let result = commandPacket.concat(valueToBytesArray(parseInt(writeValue, 10), parseInt(regInfo.size), 'little'));
    if (isCrcEnabled) {
      return result.concat(crcCalculation(result));
    }
    return result;
  }
};

function deserializeCommandPacket(thisCmd, data, isCrcEnabled) {
  if (isCrcEnabled) {
    let actualData = [];
    let crcLimit = 0;
    if (thisCmd.size == 16) {
      actualData = data.slice(5, 7);
      crcLimit = 8;
    } else if (thisCmd.size == 32) {
      actualData = data.slice(5, 9);
      crcLimit = 10;
    } else {
      actualData = data.slice(5, 13);
      crcLimit = 14;
    }
    if (crcCalculation(data.slice(0, crcLimit)) == 0) {
      return bytesArrToValue(actualData, thisCmd.size, 'little');
    } else {
      return -1;
    }
  } else {
    return bytesArrToValue(actualData, thisCmd.size, 'little');
  }
};

function crcCalculation(buffer) {
  let crcInit = 0xFF;
  let crc = 0x0;
  let i = 0,
    j = 0,
    carry = 0;

  /* Default initial value for CRC */
  crc = crcInit;

  for (i = 0; i < buffer.length; i++) {
    crc ^= buffer[i];
    /* Loop to shift the message bits one by one */
    for (j = 0; j < 8; j++) {
      carry = crc & 0x80;
      crc <<= 1;
      if (carry) {
        /* CCITT Polynomial x^8 + x^2 + x + 1 */
        crc ^= 0x07;
      }
    }
  }

  return (crc & 0xFF);
}

CustomCodec.prototype.connect = function (settings) {
  this.model = document.querySelector('ti-model-register');
  if (this.model) {
    var modelId = this.model.id;
    var refreshBind = gc.databind.registry.getBinding(modelId + '.$refresh_interval');
    if (refreshBind) {
      refreshBind.setPreRefreshInterval(this.preRefreshTimeInMs);
      this.preRefreshListener = new gc.databind.IChangedListener();
      this.preRefreshListener.onPreRefresh = this.onPreRefresh.bind(this);
      refreshBind.addPreRefreshListener(this.preRefreshListener);

      this.refreshListener = new gc.databind.IChangedListener();
      this.refreshListener.onRefresh = this.onRefresh.bind(this);
      refreshBind.addRefreshListener(this.refreshListener);

    }
    this.model.getRegisterViewInformation(0).then(function (deviceObj) {
      this.maxBitWidth = deviceObj.info.regsize;
      this.readCmdSymbols = [];
      this.regBlocks = deviceObj.regblocks;
      for (var i = 0; i < deviceObj.regblocks.length; i++) {
        var regBlock = deviceObj.regblocks[i];
        var structName = "";
        if (regBlock.structName && regBlock.structName.length > 0) {
          structName = regBlock.structName + ".";
        }
        if (regBlock.structReadCmdName) {
          var readCmdSymbol = structName + regBlock.structReadCmdName;
          this.readCmdSymbols.push(readCmdSymbol);
        }
        if (regBlock.structWriteCmdName) {
          var writeCmdSymbol = structName + regBlock.structWriteCmdName;
          this.writeCmdSymbols.push(writeCmdSymbol);
        }
      }
    }.bind(this));
  }

  gc.fileCache.readJsonFile('app/memoryAddress.json').then((content) => {
    memoryInfo = content.memoryAddress;
  })

};

// registerCustomCodec has to be called right at startup in order for the codec to be recognized and used
gc.console.log("CustomCodec", "CustomCodec: about to register...");
var baseCodecs;
gc.databind.registerCustomCodec("customRegModelCodec", CustomCodec, baseCodecs);